#include <gba.h>
#include <string.h>

#ifdef __cplusplus
extern "C" {
#endif

#define HSLtoRGB__Q(S, L) ( \
	((L) < 0.5) ? \
		((L) * (1.0 + (S))) : \
		((L) + (S) - ((L) * (S))) \
)

#define HSLtoRGB__P(S, L) (2.0 * (L) - HSLtoRGB__Q((S), (L)))

#define HSLtoRGB__T_wrap(T) ( \
	((T) < 0.0) ? ((T) + 1.0) : \
	((T) > 1.0) ? ((T) - 1.0) : (T) \
)

#define HSLtoRGB__TR(H) HSLtoRGB__T_wrap((H) + (1.0 / 3.0))

#define HSLtoRGB__TG(H) (H)

#define HSLtoRGB__TB(H) HSLtoRGB__T_wrap((H) - (1.0 / 3.0))

#define HSLtoRGB__RGB(T, S, L) ( \
	((T) < (1.0 / 6.0)) ? ( \
		HSLtoRGB__P((S), (L)) + ( \
			( \
				HSLtoRGB__Q((S), (L)) - \
				HSLtoRGB__P((S), (L)) \
			) * 6.0 * (T) \
		) \
	) : ((T) < (1.0 / 2.0)) ? HSLtoRGB__Q((S), (L)) : \
	((T) < (2.0 / 3.0)) ? ( \
		HSLtoRGB__P((S), (L)) + ( \
			( \
				HSLtoRGB__Q((S), (L)) - \
				HSLtoRGB__P((S), (L)) \
			) * 6.0 * ((2.0 / 3.0) - (T)) \
		) \
	) : HSLtoRGB__P((S), (L)) \
)

#define HSLtoRGB(H, S, L) ( \
	((int)(HSLtoRGB__RGB(HSLtoRGB__TR(H), (S), (L)) * 255.0) << 16) | \
	((int)(HSLtoRGB__RGB(HSLtoRGB__TG(H), (S), (L)) * 255.0) <<  8) | \
	((int)(HSLtoRGB__RGB(HSLtoRGB__TB(H), (S), (L)) * 255.0) <<  0) \
)

static inline int rangeChannel(u8 a, u8 b, u8 p)
	{ return (int)a + ((((int)b - (int)a) * (int)p) / 255); }

static inline u32 rangeRGB(u32 a, u32 b, u8 p) {
	return	(rangeChannel((a >> 16), (b >> 16), p) << 16) |
		(rangeChannel((a >>  8), (b >>  8), p) <<  8) |
		(rangeChannel((a >>  0), (b >>  0), p) <<  0);
}

#define RGBtoGBA(i) ( \
	(((i) >> 19) & (0xF80000 >> 19)) | \
	(((i) >>  6) & (0x00F800 >>  6)) | \
	(((i) <<  7) & (0x0000F8 <<  7)) \
)

static const u32 pipes[] = {
#include "pipes.h"
};

static const u32 pipesPal[] = {
#include "pipespal.h"
};

static const u32 pipesHues[] = {
	HSLtoRGB(( 1.0 / 32.0), 1.0, 0.5),
	HSLtoRGB(( 3.0 / 32.0), 1.0, 0.5),
	HSLtoRGB(( 5.0 / 32.0), 1.0, 0.5),
	HSLtoRGB(( 7.0 / 32.0), 1.0, 0.5),
	HSLtoRGB(( 9.0 / 32.0), 1.0, 0.5),
	HSLtoRGB((11.0 / 32.0), 1.0, 0.5),
	HSLtoRGB((13.0 / 32.0), 1.0, 0.5),
	HSLtoRGB((15.0 / 32.0), 1.0, 0.5),
	HSLtoRGB((17.0 / 32.0), 1.0, 0.5),
	HSLtoRGB((19.0 / 32.0), 1.0, 0.5),
	HSLtoRGB((21.0 / 32.0), 1.0, 0.5),
	HSLtoRGB((23.0 / 32.0), 1.0, 0.5),
	HSLtoRGB((25.0 / 32.0), 1.0, 0.5),
	HSLtoRGB((27.0 / 32.0), 1.0, 0.5),
	HSLtoRGB((29.0 / 32.0), 1.0, 0.5),
	HSLtoRGB((31.0 / 32.0), 1.0, 0.5)
};

enum { DIR_UP, DIR_RIGHT, DIR_DOWN, DIR_LEFT };
enum {
	BIT_UP = (1 << DIR_UP),
	BIT_RIGHT = (1 << DIR_RIGHT),
	BIT_DOWN = (1 << DIR_DOWN),
	BIT_LEFT = (1 << DIR_LEFT)
}; enum { BIT_ALL = (BIT_UP | BIT_RIGHT | BIT_DOWN | BIT_LEFT) };

static inline s32 updateSeed(s32 seed) { return (seed * 1103515245) + 12345; }

static inline s32 nextPos(s32 *pseed)
{ return (*pseed = updateSeed(*pseed)) & 0x7FFFFFFF; }

static inline s32 nextBits(s32 *pseed, int num)
{ return nextPos(pseed) >> (31 - num); }

static inline int rand2(s32 *pseed) { return nextBits(pseed, 1); }

static inline int rand3(s32 *pseed)
{ return nextPos(pseed) / ((0x7FFFFFFF / 3) + 1); }

static inline int rand4(s32 *pseed) { return nextBits(pseed, 2); }

static inline int rand18(s32 *pseed)
{ return nextPos(pseed) / ((0x7FFFFFFF / 18) + 1); }

static inline int rand20(s32 *pseed)
{ return nextPos(pseed) / ((0x7FFFFFFF / 20) + 1); }

static inline int rand28(s32 *pseed)
{ return nextPos(pseed) / ((0x7FFFFFFF / 28) + 1); }

static inline int rand30(s32 *pseed)
{ return nextPos(pseed) / ((0x7FFFFFFF / 30) + 1); }

static volatile int vblankCount;
static volatile s32 globalSeed;
static volatile u8 useRainbow;
static volatile u8 rainbowHue;
static u8 useSnaky;
static s8 prevX;
static s8 prevY;
static u8 useWrapX;
static u8 useWrapY;
static s32 baseSeed;
static s32 currentSeed;
static int remaining;
static volatile u16 grid[20][32];
#define GRID(x, y) grid[(y)][(x)]
static volatile u16 palettesGrey[16][16];
static volatile u16 palettesColor[32][16];

static void initPalettes(void) {
	int i, j;
	for (i = 0; i < 16; i++)
		palettesGrey[0][i] = RGBtoGBA(pipesPal[i]);
	CpuFastSet((void *)palettesGrey[0], (void *)palettesGrey[1], (16*1)>>1);
	CpuFastSet((void *)palettesGrey[0], (void *)palettesGrey[2], (16*2)>>1);
	CpuFastSet((void *)palettesGrey[0], (void *)palettesGrey[4], (16*4)>>1);
	CpuFastSet((void *)palettesGrey[0], (void *)palettesGrey[8], (16*8)>>1);
	for (j = 0; j < 16; j++) {
		for (i = 0; i < 16; i++) {
			int rgb = rangeRGB(0x000000, pipesHues[j], pipesPal[i]);
			palettesColor[j][i] = RGBtoGBA(rgb);
		}
	}
	CpuFastSet((void *)palettesColor[0], (void *)palettesColor[16],
		(16*16)>>1);
}

void mainInterrupt(void) {
	int count = vblankCount++;
	CpuFastSet(grid, SCREEN_BASE_BLOCK(1), (20*32)>>1);
	if ((count & 3) == 0) {
		int hue = rainbowHue++ & 0xF;
		if (useRainbow)
			CpuFastSet((void *)palettesColor[hue],
				BG_PALETTE, 256>>1);
		else
			CpuFastSet((void *)palettesGrey,
				BG_PALETTE, 256>>1);
	}
	globalSeed = updateSeed(globalSeed);
}

static inline int wrapX(x)
	{ return (x < 0) ? (x + 30) : (x >= 30) ? (x - 30) : x; }
static inline int wrapY(y)
	{ return (y < 0) ? (y + 20) : (y >= 20) ? (y - 20) : y; }

static void resetMaze(void) {
	int x, y;
	memset(grid, 0, 20*32*2);
	remaining = 20*30;
	currentSeed = baseSeed;
	x = !useWrapX ? (rand28(&currentSeed) + 1) : rand30(&currentSeed);
	y = !useWrapY ? (rand18(&currentSeed) + 1) : rand20(&currentSeed);
	switch ((u8)rand4(&currentSeed)) {
	case DIR_UP:
		GRID(x, y) = BIT_UP;
		y = wrapY(y - 1);
		GRID(x, y) = BIT_DOWN | (-1 << 12);
		break;
	case DIR_RIGHT:
		GRID(x, y) = BIT_RIGHT;
		x = wrapX(x + 1);
		GRID(x, y) = BIT_LEFT | (-1 << 12);
		break;
	case DIR_DOWN:
		GRID(x, y) = BIT_DOWN;
		y = wrapY(y + 1);
		GRID(x, y) = BIT_UP | (-1 << 12);
		break;
	case DIR_LEFT:
		GRID(x, y) = BIT_LEFT;
		x = wrapX(x - 1);
		GRID(x, y) = BIT_RIGHT | (-1 << 12);
	}
	prevX = x;
	prevY = y;
	remaining -= 2;
}

static void newMaze(void) {
	baseSeed = globalSeed;
	resetMaze();
}

static void newBranch(void) {
	do {
		prevX = rand30(&currentSeed);
		prevY = rand20(&currentSeed);
	} while (!(GRID(prevX, prevY) & BIT_ALL));
}

static void iterateMaze(void) {
	u8 dir = 0;
	if (!useSnaky)
		newBranch();
	for (;;) {
		s8 x, y, xl, xr, yu, yd;
		u8 cu, cr, cd, cl;
		u16 value;
		x = prevX;
		y = prevY;
		xl = wrapX(x - 1);
		xr = wrapX(x + 1);
		yu = wrapY(y - 1);
		yd = wrapY(y + 1);
		cu = !((!useWrapY && y == 0) || (GRID(x, yu) & BIT_ALL));
		cr = !((!useWrapX && x == 29) || (GRID(xr, y) & BIT_ALL));
		cd = !((!useWrapY && y == 19) || (GRID(x, yd) & BIT_ALL));
		cl = !((!useWrapX && x == 0) || (GRID(xl, y) & BIT_ALL));
if (cu) {
	if (cr) {
		if (cd) {
			if (cl) {
				dir = rand4(&currentSeed);
			} else { /* !cl */
				static const u8 DIRS[] =
					{ DIR_UP, DIR_RIGHT,
						DIR_RIGHT, DIR_DOWN };
				dir = DIRS[rand4(&currentSeed)];
			}
		} else { /* !cd */
			if (cl) {
				static const u8 DIRS[] =
					{ DIR_UP, DIR_UP, DIR_RIGHT, DIR_LEFT };
				dir = DIRS[rand4(&currentSeed)];
			} else { /* !cl */
				static const u8 DIRS[] =
					{ DIR_UP, DIR_RIGHT };
				dir = DIRS[rand2(&currentSeed)];
			}
		}
	} else { /* !cr */
		if (cd) {
			if (cl) {
				static const u8 DIRS[] =
					{ DIR_UP, DIR_DOWN,
						DIR_LEFT, DIR_LEFT };
				dir = DIRS[rand4(&currentSeed)];
			} else { /* !cl */
				static const u8 DIRS[] =
					{ DIR_UP, DIR_DOWN };
				dir = DIRS[rand2(&currentSeed)];
			}
		} else { /* !cd */
			if (cl) {
				static const u8 DIRS[] =
					{ DIR_UP, DIR_LEFT };
				dir = DIRS[rand2(&currentSeed)];
			} else { /* !cl */
				dir = DIR_UP;
			}
		}
	}
} else { /* !cu */
	if (cr) {
		if (cd) {
			if (cl) {
				static const u8 DIRS[] =
					{ DIR_RIGHT, DIR_DOWN,
						DIR_LEFT, DIR_DOWN };
				dir = DIRS[rand4(&currentSeed)];
			} else { /* !cl */
				static const u8 DIRS[] =
					{ DIR_RIGHT, DIR_DOWN };
				dir = DIRS[rand2(&currentSeed)];
			}
		} else { /* !cd */
			if (cl) {
				static const u8 DIRS[] =
					{ DIR_RIGHT, DIR_LEFT };
				dir = DIRS[rand2(&currentSeed)];
			} else { /* !cl */
				dir = DIR_RIGHT;
			}
		}
	} else { /* !cr */
		if (cd) {
			if (cl) {
				static const u8 DIRS[] =
					{ DIR_DOWN, DIR_LEFT };
				dir = DIRS[rand2(&currentSeed)];
			} else { /* !cl */
				dir = DIR_DOWN;
			}
		} else { /* !cd */
			if (cl) {
				dir = DIR_LEFT;
			} else { /* !cl */
				newBranch();
				continue;
			}
		}
	}
}
		switch (dir) {
		case DIR_UP:
			value = (GRID(x, y) |= BIT_UP) & (0xF << 12);
			y = yu;
			GRID(x, y) = BIT_DOWN | (value - (1 << 12));
			break;
		case DIR_RIGHT:
			value = (GRID(x, y) |= BIT_RIGHT) & (0xF << 12);
			x = xr;
			GRID(x, y) = BIT_LEFT | (value - (1 << 12));
			break;
		case DIR_DOWN:
			value = (GRID(x, y) |= BIT_DOWN) & (0xF << 12);
			y = yd;
			GRID(x, y) = BIT_UP | (value - (1 << 12));
			break;
		case DIR_LEFT:
			value = (GRID(x, y) |= BIT_LEFT) & (0xF << 12);
			x = xl;
			GRID(x, y) = BIT_RIGHT | (value - (1 << 12));
			break;
		}
		prevX = x;
		prevY = y;
		remaining--;
		break;
	}
}

static void handleInput(void) {
	u16 keys;
	scanKeys();
	keys = keysDown();
	if (keys & KEY_A) {
		useSnaky = 1;
	} else if (keys & KEY_B) {
		useSnaky = 0;
	} else if (keys & KEY_SELECT) {
		useRainbow = 0;
	} else if (keys & KEY_START) {
		useRainbow = 1;
	} else if (keys & (KEY_RIGHT | KEY_LEFT)) {
		useWrapX = !useWrapX;
	} else if (keys & (KEY_UP | KEY_DOWN)) {
		useWrapY = !useWrapY;
	} else if (keys & KEY_R) {
		if (remaining) {
			do { iterateMaze(); } while (remaining);
		} else {
			newMaze();
		}
	} else if (keys & KEY_L) {
		resetMaze();
	}
}

int main(void) {
	vblankCount = 0;
	useRainbow = 1;
	rainbowHue = 1;
	useSnaky = 1;
	useWrapX = useWrapY = 0;
	initPalettes();
	newMaze();
	irqInit();
	irqSet(IRQ_VBLANK, mainInterrupt);
	irqEnable(IRQ_VBLANK);
	REG_IME = 1;
	REG_KEYCNT = KEY_A | KEY_B | KEY_SELECT | KEY_START |
		DPAD | KEY_R | KEY_L;
	REG_DISPCNT |= 1 << 7;
	REG_DISPCNT &= ~7;
	REG_DISPCNT |= 1 << 8;
	REG_DISPCNT &= ~(0x7F << 9);
	REG_BG0CNT = 1 << 8;
	REG_BG0HOFS = 0;
	REG_BG0VOFS = 0;
	CpuFastSet(pipes, (void *)VRAM, 16<<3);
	VBlankIntrWait();
	REG_DISPCNT &= ~(1 << 7);
	for (;;) {
		handleInput();
		if (remaining && ((vblankCount & 3) == 0))
			iterateMaze();
		VBlankIntrWait();
	} return 0;
}

#ifdef __cplusplus
} /* end extern "C" */
#endif